from typing import Optional, List

from dynatrace.dynatrace_object import DynatraceObject
from dynatrace.http_client import HttpClient

UNIT_BIT = "Bit(bit)"
UNIT_BITPERHOUR = "BitPerHour(bit/h)"
UNIT_BITPERMINUTE = "BitPerMinute(bit/min)"
UNIT_BITPERSECOND = "BitPerSecond(bit/s)"
UNIT_BYTE = "Byte(B)"
UNIT_BYTEPERHOUR = "BytePerHour(B/h)"
UNIT_BYTEPERMINUTE = "BytePerMinute(B/min)"
UNIT_BYTEPERSECOND = "BytePerSecond(B/s)"
UNIT_CORES = "Cores"
UNIT_COUNT = "Count(count)"
UNIT_DAY = "Day(ds)"
UNIT_DECIBELMILLIWATT = "DecibelMilliWatt(dBm)"
UNIT_G = "G"
UNIT_GIBIBYTE = "GibiByte(GiB)"
UNIT_GIGABYTE = "GigaByte(GB)"
UNIT_HOUR = "Hour(hs)"
UNIT_KIBIBYTE = "KibiByte(KiB)"
UNIT_KIBIBYTEPERHOUR = "KibiBytePerHour(KiB/h)"
UNIT_KIBIBYTEPERMINUTE = "KibiBytePerMinute(KiB/min)"
UNIT_KIBIBYTEPERSECOND = "KibiBytePerSecond(KiB/s)"
UNIT_KILOBYTE = "KiloByte(kB)"
UNIT_KILOBYTEPERHOUR = "KiloBytePerHour(kB/h)"
UNIT_KILOBYTEPERMINUTE = "KiloBytePerMinute(kB/min)"
UNIT_KILOBYTEPERSECOND = "KiloBytePerSecond(kB/s)"
UNIT_M = "M"
UNIT_MEBIBYTE = "MebiByte(MiB)"
UNIT_MEBIBYTEPERHOUR = "MebiBytePerHour(MiB/h)"
UNIT_MEBIBYTEPERMINUTE = "MebiBytePerMinute(MiB/min)"
UNIT_MEBIBYTEPERSECOND = "MebiBytePerSecond(MiB/s)"
UNIT_MEGABYTE = "MegaByte(MB)"
UNIT_MEGABYTEPERHOUR = "MegaBytePerHour(MB/h)"
UNIT_MEGABYTEPERMINUTE = "MegaBytePerMinute(MB/min)"
UNIT_MEGABYTEPERSECOND = "MegaBytePerSecond(MB/s)"
UNIT_MICROSECOND = "MicroSecond(µs)"
UNIT_MILLISECOND = "MilliSecond(ms)"
UNIT_MILLISECONDPERMINUTE = "MilliSecondPerMinute(ms/min)"
UNIT_MINUTE = "Minute(mins)"
UNIT_MONTH = "Month(mos)"
UNIT_NOT_APPLICABLE = "N/A"
UNIT_NANOSECOND = "NanoSecond(ns)"
UNIT_NANOSECONDPERMINUTE = "NanoSecondPerMinute(ns/min)"
UNIT_PERHOUR = "PerHour(count/h)"
UNIT_PERMINUTE = "PerMinute(count/min)"
UNIT_PERSECOND = "PerSecond(count/s)"
UNIT_PERCENT = "Percent(%)"
UNIT_PIXEL = "Pixel(px)"
UNIT_PROMILLE = "Promille(‰)"
UNIT_RATIO = "Ratio"
UNIT_SECOND = "Second(s)"
UNIT_STATE = "State"
UNIT_UNSPECIFIED = "Unspecified"
UNIT_WEEK = "Week(ws)"
UNIT_YEAR = "Year(ys)"
UNIT_K = "k"
UNIT_M_CORES = "mCores"


class TimeSerieService:
    def __init__(self, http_client: HttpClient):
        self.__http_client = http_client

    def create_timeseries(
        self,
        metric_id: str,
        display_name: Optional[str],
        unit: Optional[str] = None,
        dimensions: Optional[List[str]] = None,
        technologies: Optional[List[str]] = None,
    ) -> "TimeseriesRegistrationMessage":

        return TimeseriesRegistrationMessage(
            self.__http_client,
            metric_id=metric_id,
            display_name=display_name,
            unit=unit,
            dimensions=dimensions,
            technologies=technologies,
        )


class TimeseriesRegistrationMessage(DynatraceObject):
    def __init__(
        self,
        http_client,
        metric_id: str,
        display_name: Optional[str],
        unit: Optional[str] = None,
        dimensions: Optional[List[str]] = None,
        technologies: Optional[List[str]] = None,
    ):
        self.metric_id = metric_id

        raw_element = {
            "displayName": display_name,
            "unit": unit,
            "dimensions": dimensions,
            "types": technologies,
        }
        super().__init__(http_client, None, raw_element)

    def put(self):
        return self._http_client.make_request(f"/api/v1/timeseries/{self.metric_id}", params=self._raw_element, method="PUT")



from datetime import datetime
from typing import List, Optional

from dynatrace.environment_v2.entity_type import EntityType
from dynatrace.http_client import HttpClient
from dynatrace.configuration_v1.management_zone import ManagementZone
from dynatrace.configuration_v1.metag import METag
from dynatrace.dynatrace_object import DynatraceObject
from dynatrace.pagination import PaginatedList


class EntityService:
    def __init__(self, http_client: HttpClient):
        self.__http_client = http_client

    def list(
        self,
        entity_selector: str,
        time_from: str = "now-2h",
        time_to: str = "now",
        fields: Optional[str] = None,
        page_size=50,
    ) -> PaginatedList["Entity"]:
        """
        :return: A list of monitored entities along with their properties.
        """
        params = {"pageSize": page_size, "entitySelector": entity_selector, "from": time_from, "to": time_to, "fields": fields}
        return PaginatedList(Entity, self.__http_client, "/api/v2/entities", params, list_item="entities")

    def list_types(self, page_size=50) -> PaginatedList[EntityType]:
        """
        Gets a list of properties for all entity types

        :param page_size: The desired amount of entities in a single response payload.
            The maximal allowed page size is 500.
            If not set, 50 is used.
        :return: A list of properties of all available entity types.
        """
        params = {"pageSize": page_size}
        return PaginatedList("EntityType", self.__http_client, "/api/v2/entityTypes", params, list_item="types")


class Entity(DynatraceObject):
    @property
    def from_relationships(self):
        # TODO
        return ""

    @property
    def to_relationships(self):
        # TODO
        return ""

    @property
    def first_seen_t_ms(self) -> datetime:
        # TODO
        """
        The timestamp at which the entity was first seen, in UTC milliseconds.
        :return:
        """
        return datetime.now()

    @property
    def last_seen_t_ms(self) -> datetime:
        # TODO
        """
        The timestamp at which the entity was last seen, in UTC milliseconds.
        :return:
        """
        return datetime.now()

    @property
    def entity_id(self) -> str:
        """
        The ID of the entity.
        :return:
        """
        return self._entity_id

    @property
    def display_name(self) -> str:
        """
        The name of the entity, displayed in the UI.
        :return:
        """
        return self._display_name

    @property
    def management_zones(self) -> List[ManagementZone]:
        return []

    @property
    def tags(self) -> List[METag]:
        return []

    @property
    def properties(self) -> List[dict]:
        return self._properties

    def _create_from_raw_data(self, raw_element: dict):
        self._display_name = raw_element.get("displayName")
        self._entity_id = raw_element.get("entityId")
        self._properties = raw_element.get("properties", {})


class EntityShortRepresentation(DynatraceObject):
    def _create_from_raw_data(self, raw_element):
        self.id = raw_element.get("id")
        self.name = raw_element.get("name")
        self.description = raw_element.get("description")



from requests import Response
from dynatrace.environment_v2.entity import EntityShortRepresentation


class EndpointShortRepresentation(EntityShortRepresentation):
    def delete(self, plugin_id: str) -> Response:
        """
        Deletes this endpoint
        """
        return self._http_client.make_request(f"/api/config/v1/plugins/{plugin_id}/endpoints/{self.id}", method="DELETE")



from typing import List

from dynatrace.configuration_v1.management_zone import ManagementZone
from dynatrace.configuration_v1.metag import METag
from dynatrace.dynatrace_object import DynatraceObject


class EntityType(DynatraceObject):
    @property
    def from_relationships(self):
        return ""

    @property
    def to_relationships(self):
        return ""

    @property
    def type(self) -> str:
        """
        The name of the entity, displayed in the UI.
        :return:
        """
        return self._entity_type

    @property
    def management_zones(self) -> List[ManagementZone]:
        return []

    @property
    def tags(self) -> List[METag]:
        return []

    @property
    def properties(self) -> List[dict]:
        return self._properties

    def _create_from_raw_data(self, raw_element: dict):
        self._entity_type = raw_element.get("type")
        self._properties = raw_element.get("properties", [])



import pprint
from typing import Optional, Dict, Any

from requests import Response

from dynatrace.http_client import HttpClient


class DynatraceObject:
    def __init__(self, http_client: Optional[HttpClient] = None, headers: Optional[Dict[str, str]] = None, raw_element: Optional[Dict[str, Any]] = None):
        if raw_element is None:
            raw_element = {}
        self._http_client = http_client
        self._headers = headers
        self._raw_element = raw_element
        self._create_from_raw_data(raw_element)

    def _create_from_raw_data(self, raw_element: Dict[str, Any]):
        pass

    def __repr__(self):
        return f"{self.__class__.__name__}({pprint.pformat(self._raw_element, width=130)})"

    def _make_request(self, path: str, params: Optional[Dict] = None, headers: Optional[Dict] = None, method="GET", data=None) -> Response:
        return self._http_client.make_request(path, params, headers, method, data)



from typing import List
from dynatrace.dynatrace_object import DynatraceObject


class ConfigurationMetadata(DynatraceObject):
    def _create_from_raw_data(self, raw_element):
        self.cluster_version: str = raw_element.get("clusterVersion")
        self.configuration_versions: List[int] = raw_element.get("configurationVersions")
        self.current_configuration_versions: List[int] = raw_element.get("currentConfigurationVersions")



from datetime import datetime

from dynatrace.http_client import HttpClient


class ClusterTimeService:
    def __init__(self, http_client: HttpClient):
        self.__http_client = http_client

    def time(self) -> datetime:
        return datetime.utcfromtimestamp(float(self.__http_client.make_request("/api/v1/time").text) / 1000)



from typing import List

from requests import Response

from dynatrace.environment_v2.configuration import ConfigurationMetadata
from dynatrace.dynatrace_object import DynatraceObject
from dynatrace.configuration_v1.endpoint import EndpointShortRepresentation
from dynatrace.environment_v2.entity import EntityShortRepresentation
from dynatrace.http_client import HttpClient
from dynatrace.pagination import PaginatedList


class ExtensionService:
    def __init__(self, http_client: HttpClient):
        self.__http_client = http_client
        pass

    def list_endpoints(self, plugin_id: str) -> PaginatedList[EndpointShortRepresentation]:
        """
        Lists endpoints of the specified ActiveGate plugin
        """
        return PaginatedList("EndpointShortRepresentation", self.__http_client, f"/api/config/v1/plugins/{plugin_id}/endpoints", list_item="values")

    def list(self, page_size: int = 200) -> PaginatedList["ExtensionDto"]:
        """
        List all uploaded extensions

        :param page_size: The number of results per result page. Must be between 1 and 500
            Default value : 200
        """
        params = {"pageSize": page_size}
        return PaginatedList(ExtensionDto, self.__http_client, f"/api/config/v1/extensions", params, list_item="extensions")

    def get(self, extension_id: str):
        response = self.__http_client.make_request(f"/api/config/v1/extensions/{extension_id}").json()
        return Extension(self.__http_client, None, response)

    def list_instances(self, extension_id: str, page_size: int = 200) -> PaginatedList["ExtensionShortRepresentation"]:
        params = {"pageSize": page_size}
        return PaginatedList(
            ExtensionShortRepresentation,
            self.__http_client,
            f"/api/config/v1/extensions/{extension_id}/instances",
            list_item="configurationsList",
            target_params=params,
        )

    def get_instance(self, extension_id: str, configuration_id: str):
        response = self.__http_client.make_request(f"/api/config/v1/extensions/{extension_id}/instances/{configuration_id}").json()
        return ExtensionConfigurationDto(self.__http_client, None, response)

    def post_instance(self, extension_configuration: "ExtensionConfigurationDto"):
        return extension_configuration.post()

    def create_instance(
        self,
        extension_id: str,
        enabled: bool = True,
        use_global: bool = True,
        properties: dict = None,
        host_id: str = None,
        active_gate: EntityShortRepresentation = None,
        endpoint_id: str = None,
        endpoint_name: str = None,
    ) -> "ExtensionConfigurationDto":

        return ExtensionConfigurationDto(self.__http_client, extension_id, enabled, use_global, properties, host_id, active_gate, endpoint_id, endpoint_name)


class ExtensionProperty(DynatraceObject):
    @property
    def key(self) -> str:
        return self._key

    @property
    def type(self) -> str:
        return self._type

    @property
    def default_value(self) -> str:
        return self._default_value

    @property
    def dropdown_values(self) -> List[str]:
        return self._dropdown_values

    def _create_from_raw_data(self, raw_element):
        self._key = raw_element.get("key")
        self._type = raw_element.get("type")
        self._default_value = raw_element.get("defaultValue")
        self._dropdown_values = raw_element.get("dropdownValues")


class ExtensionConfigurationDto(DynatraceObject):
    def __init__(
        self,
        http_client,
        extension_id: str,
        enabled: bool = True,
        use_global: bool = True,
        properties: dict = None,
        host_id: str = None,
        active_gate: EntityShortRepresentation = None,
        endpoint_id: str = None,
        endpoint_name: str = None,
    ):
        raw_element = {
            "extensionId": extension_id,
            "enabled": enabled,
            "useGlobal": use_global,
            "properties": properties,
            "hostId": host_id,
            "activeGate": active_gate._raw_element,
            "endpointId": endpoint_id,
            "endpointName": endpoint_name,
        }

        super().__init__(http_client, None, raw_element)

    def post(self):
        return self._http_client.make_request(f"/api/config/v1/extensions/{self.extension_id}/instances", params=self._raw_element, method="POST")

    @property
    def extension_id(self) -> str:
        return self._extension_id

    @property
    def enabled(self) -> bool:
        return self._enabled

    @property
    def use_global(self) -> bool:
        return self._use_global

    @property
    def properties(self) -> dict:
        return self._properties

    @property
    def host_id(self) -> str:
        return self._host_id

    @property
    def active_gate(self) -> EntityShortRepresentation:
        return self._active_gate

    @property
    def endpoint_id(self) -> str:
        return self._endpoint_id

    @property
    def endpoint_name(self) -> str:
        return self._endpoint_name

    def _create_from_raw_data(self, raw_element):
        self._extension_id = raw_element.get("extensionId")
        self._enabled = raw_element.get("enabled")
        self._use_global = raw_element.get("useGlobal")
        self._properties = raw_element.get("properties")
        self._host_id = raw_element.get("hostId")
        self._active_gate = EntityShortRepresentation(self._http_client, None, raw_element.get("activeGate"))
        self._endpoint_id = raw_element.get("endpointId")
        self._endpoint_name = raw_element.get("endpointName")


class ExtensionShortRepresentation(EntityShortRepresentation):
    def get_full_configuration(self, extension_id: str):
        """
        Gets the full extension configuration for this ExtensionShortRepresentation
        """
        response = self._http_client.make_request(f"/api/config/v1/extensions/{extension_id}/instances/{self.id}").json()
        return ExtensionConfigurationDto(self._http_client, None, response)


class Extension(DynatraceObject):
    @property
    def id(self) -> str:
        return self._id

    @property
    def name(self) -> str:
        return self._name

    @property
    def type(self) -> str:
        return self._type

    @property
    def version(self):
        return self._version

    @property
    def metric_group(self):
        return self._metric_group

    @property
    def metadata(self) -> ConfigurationMetadata:
        return self._metadata

    @property
    def properties(self) -> List[ExtensionProperty]:
        return self._properties

    def _create_from_raw_data(self, raw_element):
        self._id = raw_element.get("id")
        self._name = raw_element.get("name")
        self._type = raw_element.get("type")
        self._version = raw_element.get("version")
        self._metric_group = raw_element.get("metricGroup")
        self._metadata = ConfigurationMetadata(self._http_client, None, raw_element.get("metadata"))
        self._properties = [ExtensionProperty(self._http_client, None, prop) for prop in raw_element.get("properties")]


class ExtensionDto(DynatraceObject):
    @property
    def id(self) -> str:
        return self._id

    @property
    def name(self):
        return self._name

    @property
    def type(self):
        return self._type

    def _create_from_raw_data(self, raw_element):
        self._id = raw_element.get("id")
        self._name = raw_element.get("name")
        self._type = raw_element.get("type")

    def get_full_extension(self) -> Extension:
        """
        Gets the full extension for this ExtensionDto
        """
        response = self._http_client.make_request(f"/api/config/v1/extensions/{self.id}").json()
        return Extension(self._http_client, None, response)

    @property
    def instances(self) -> PaginatedList[ExtensionShortRepresentation]:
        return PaginatedList(
            ExtensionShortRepresentation,
            self._http_client,
            f"/api/config/v1/extensions/{self.id}/instances",
            list_item="configurationsList",
        )

    def delete(self) -> Response:
        """
        Deletes the ZIP file of this extension
        """
        return self._http_client.make_request(f"/api/config/v1/extensions/{self.id}", method="DELETE")



from datetime import datetime, timedelta, timezone
from collections import MutableSequence
from typing import Optional, List, Dict, Tuple


from dynatrace.dynatrace_object import DynatraceObject
from dynatrace.http_client import HttpClient


class CustomDeviceService:
    def __init__(self, http_client: HttpClient):
        self.__http_client = http_client

    def create(
        self,
        device_id: str,
        display_name: Optional[str] = None,
        group: Optional[str] = None,
        ip_addresses: Optional[List[str]] = None,
        listen_ports: Optional[List[int]] = None,
        technology: Optional[str] = None,
        favicon: Optional[str] = None,
        config_url: Optional[str] = None,
        properties: Optional[Dict[str, str]] = None,
        tags: Optional[List[str]] = None,
        series: Optional[List] = None,
        host_names: Optional[List[str]] = None,
    ) -> "CustomDevicePushMessage":
        return CustomDevicePushMessage(
            self.__http_client,
            device_id=device_id,
            display_name=display_name,
            group=group,
            ip_addresses=ip_addresses,
            listen_ports=listen_ports,
            technology=technology,
            favicon=favicon,
            config_url=config_url,
            properties=properties,
            tags=tags,
            series=series,
            host_names=host_names,
        )


class Series(MutableSequence):
    def __init__(self, *args):
        self.list: List["EntityTimeseriesData"] = []
        self.extend(list(args))

    def append(self, time_series: "EntityTimeseriesData") -> None:
        for element in self.list:
            if time_series.timeseries_id == element.timeseries_id and time_series.dimensions == element.dimensions:
                element.data_points.extend(time_series.data_points)
                return
        self.list.append(time_series)

    def __len__(self):
        return len(self.list)

    def __getitem__(self, i):
        return self.list[i]

    def __delitem__(self, i):
        del self.list[i]

    def __setitem__(self, i, v):
        self.list[i] = v

    def insert(self, i, v):
        self.list.insert(i, v)

    def __str__(self):
        return str(self.list)


class CustomDevicePushMessage(DynatraceObject):
    def __init__(
        self,
        http_client,
        device_id: str,
        display_name: Optional[str] = None,
        group: Optional[str] = None,
        ip_addresses: Optional[List[str]] = None,
        listen_ports: Optional[List[int]] = None,
        technology: Optional[str] = None,
        favicon: Optional[str] = None,
        config_url: Optional[str] = None,
        properties: Optional[Dict[str, str]] = None,
        tags: Optional[List[str]] = None,
        series: Optional[Series] = None,
        host_names: Optional[List[str]] = None,
    ):
        self.device_id = device_id
        self.display_name: Optional[str] = display_name
        self.group: Optional[str] = group
        self.ip_addresses: Optional[List[str]] = ip_addresses
        self.listen_ports: Optional[List[int]] = listen_ports
        self.technology: Optional[str] = technology
        self.favicon: Optional[str] = favicon
        self.config_url: Optional[str] = config_url
        self.properties: Optional[Dict[str, str]] = properties
        self.tags: Optional[List[str]] = tags
        self.__series: Series = series
        if self.__series is None:
            self.__series: Series = Series()
        self.host_names: Optional[List[str]] = host_names

        raw_element = {
            "displayName": self.display_name,
            "group": self.group,
            "ipAddresses": self.ip_addresses,
            "listenPorts": self.listen_ports,
            "type": self.technology,
            "favicon": self.favicon,
            "configUrl": self.config_url,
            "properties": self.properties,
            "tags": self.tags,
            "series": [s._raw_element for s in self.__series] if self.__series else None,
            "hostNames": self.host_names,
        }
        super().__init__(http_client, None, raw_element)

    @property
    def series(self) -> Series:
        return self.__series

    @series.setter
    def series(self, series: Series):
        self.__series = series
        self._raw_element["series"] = [s._raw_element for s in self.__series]

    def post(self, only_valid_data_points=False):
        try:
            response = self._http_client.make_request(f"/api/v1/entity/infrastructure/custom/{self.device_id}", params=self._raw_element, method="POST")
            return response
        except Exception as e:
            if only_valid_data_points and ("configuration.Creation timestamp is:" in f"{e}" or "Data point timestamp is too far in the past" in f"{e}"):
                if "configuration.Creation timestamp" in f"{e}":
                    max_timestamp = int(f"{e}".split("configuration.Creation timestamp is:")[1].split('"')[0].strip())
                    max_timestamp = datetime.fromtimestamp(max_timestamp / 1000, tz=timezone.utc)
                else:
                    max_timestamp = datetime.now(tz=timezone.utc) - timedelta(minutes=59)
                self._http_client.log.warning(f"Some data points were invalid, removing data points older than {max_timestamp}")
                for s in self.series:
                    s.data_points = [d for d in s.data_points if d.timestamp.replace(tzinfo=max_timestamp.tzinfo) >= max_timestamp]
                self._raw_element["series"] = [s._raw_element for s in self.series]
                return self.post()
            else:
                raise e

    def absolute(self, key: str, value: float, timestamp: Optional[datetime] = None, dimensions: Optional[Dict[str, str]] = None):
        data_point = DataPoint(value, timestamp)
        self.series.append(EntityTimeseriesData(self._http_client, key, [data_point], dimensions))
        self.series = self.series  # Ugly as hell hack because of setter, and I don't want to subclass list


class EntityTimeseriesData(DynatraceObject):
    def __init__(self, http_client, timeseries_id: str, data_points: List["DataPoint"], dimensions: Optional[Dict[str, str]] = None):
        self.timeseries_id: str = timeseries_id
        self.dimensions = dimensions
        self.__data_points: List["DataPoint"] = data_points
        raw_element = {
            "timeseriesId": timeseries_id,
            "dimensions": dimensions,
            "dataPoints": [[int(data_point.timestamp.timestamp() * 1000), data_point.value] for data_point in data_points],
        }
        super().__init__(http_client, None, raw_element)

    def __repr__(self):
        return f"Series(id={self.timeseries_id}, data_points={self.data_points})"

    @property
    def data_points(self) -> List["DataPoint"]:
        return self.__data_points

    @data_points.setter
    def data_points(self, data_points: List["DataPoint"]):
        self.__data_points = data_points
        self._raw_element["dataPoints"] = [[int(data_point.timestamp.timestamp() * 1000), data_point.value] for data_point in self.__data_points]


class DataPoint:
    def __init__(self, value: float, timestamp: Optional[datetime] = None):
        self.timestamp = timestamp
        if self.timestamp is None:
            self.timestamp = datetime.now()
        self.value = value

    def __repr__(self):
        return f"[{self.timestamp}, {self.value}]"



from datetime import datetime
from typing import Optional, List

from requests import Response

from dynatrace.dynatrace_object import DynatraceObject
from dynatrace.http_client import HttpClient

EVENT_TYPE_AVAILABILITY_EVENT = "AVAILABILITY_EVENT"
EVENT_TYPE_CUSTOM_ALERT = "CUSTOM_ALERT"
EVENT_TYPE_CUSTOM_ANNOTATION = "CUSTOM_ANNOTATION"
EVENT_TYPE_CUSTOM_CONFIGURATION = "CUSTOM_CONFIGURATION"
EVENT_TYPE_CUSTOM_DEPLOYMENT = "CUSTOM_DEPLOYMENT"
EVENT_TYPE_CUSTOM_INFO = "CUSTOM_INFO"
EVENT_TYPE_ERROR_EVENT = "ERROR_EVENT"
EVENT_TYPE_MARKED_FOR_TERMINATION = "MARKED_FOR_TERMINATION"
EVENT_TYPE_PERFORMANCE_EVENT = "PERFORMANCE_EVENT"
EVENT_TYPE_RESOURCE_CONTENTION = "RESOURCE_CONTENTION"


class EventService:
    def __init__(self, http_client: HttpClient):
        self.__http_client = http_client

    def create_event(
        self,
        event_type: str,
        entity_id: str,
        source: str,
        start: Optional[datetime] = None,
        end: Optional[datetime] = None,
        timeout_minutes: Optional[int] = None,
        annotation_type: Optional[str] = None,
        annotation_description: Optional[str] = None,
        description: Optional[str] = None,
        title: Optional[str] = None,
        custom_properties: Optional[str] = None,
        allow_davis_merge: Optional[bool] = None,
    ) -> Response:

        attach_rules = PushEventAttachRules(entity_ids=[entity_id], tag_rule=None)
        return EventCreation(
            self.__http_client,
            event_type=event_type,
            attach_rules=attach_rules,
            source=source,
            start=start,
            end=end,
            timeout_minutes=timeout_minutes,
            annotation_type=annotation_type,
            annotation_description=annotation_description,
            description=description,
            title=title,
            custom_properties=custom_properties,
            allow_davis_merge=allow_davis_merge,
        ).post()


class EventCreation(DynatraceObject):
    def __init__(
        self,
        http_client,
        event_type: str,
        attach_rules: "PushEventAttachRules",
        source: str,
        start: Optional[datetime] = None,
        end: Optional[datetime] = None,
        timeout_minutes: Optional[int] = None,
        annotation_type: Optional[str] = None,
        annotation_description: Optional[str] = None,
        description: Optional[str] = None,
        title: Optional[str] = None,
        custom_properties: Optional[str] = None,
        allow_davis_merge: Optional[bool] = None,
    ):

        raw_element = {
            "eventType": event_type,
            "start": int(start.timestamp()) * 1000 if start else None,
            "end": int(end.timestamp()) * 1000 if start else None,
            "timeoutMinutes": timeout_minutes,
            "source": source,
            "annotationType": annotation_type,
            "annotationDescription": annotation_description,
            "attachRules": attach_rules._raw_element,
            "description": description,
            "title": title,
            "customProperties": custom_properties,
            "allowDavisMerge": allow_davis_merge,
        }

        super().__init__(http_client, None, raw_element)

    def post(self):
        return self._http_client.make_request(f"/api/v1/events", params=self._raw_element, method="POST")


class PushEventAttachRules:
    def __init__(self, entity_ids: Optional[List[str]], tag_rule: Optional[List["TagMatchRule"]]):

        self._raw_element = {
            "entityIds": entity_ids,
            "tagRule": [t._raw_element for t in tag_rule] if tag_rule else None,
        }


class TagMatchRule:
    def __init__(self, me_types: List[str], tags: List[str]):
        self._raw_element = {
            "meTypes": me_types,
            "tags": tags,
        }


"""
type EventStoreResult struct {
	StoredEventIds       []int    `json:"storedEventIds,omitempty"`
	StoredIds            []string `json:"storedIds,omitempty"`
	StoredCorrelationIds []string `json:"storedCorrelationIds,omitempty"`
}
"""



from typing import Generic, TypeVar, Iterator, TYPE_CHECKING

from dynatrace.dynatrace_object import DynatraceObject
from dynatrace.http_client import HttpClient

T = TypeVar("T", bound=DynatraceObject)


class PaginatedList(Generic[T]):
    def __init__(self, target_class, http_client, target_url, target_params=None, headers=None, list_item="result"):
        self.__elements = list()
        self.__target_class = target_class
        self.__http_client: HttpClient = http_client
        self.__target_url = target_url
        self.__target_params = target_params
        self.__headers = headers
        self.__list_item = list_item
        self._has_next_page = True
        self.__total_count = None
        self.__page_size = None

    def __getitem__(self, index):
        pass

    def __iter__(self) -> Iterator[T]:
        for element in self.__elements:
            yield element

        while self._has_next_page:
            new_elements = self._get_next_page()
            for element in new_elements:
                yield element

    def __len__(self):
        return self.__total_count or len(self.__elements)

    def _get_next_page(self):
        response = self.__http_client.make_request(self.__target_url, self.__target_params, self.__headers)
        json_response = response.json()
        data = []
        if json_response.get("nextPageKey", None):
            self._has_next_page = True
            self.__target_params = {"nextPageKey": json_response["nextPageKey"]}
        else:
            self._has_next_page = False

        if self.__list_item in json_response:
            elements = json_response[self.__list_item]
            self.__total_count = json_response.get("totalCount") or len(elements)

            data = [self.__target_class(self.__http_client, response.headers, element) for element in elements]
        return data



from typing import Union, Optional, List
from datetime import datetime

from requests import Response

from dynatrace.dynatrace_object import DynatraceObject
from dynatrace.http_client import HttpClient
from dynatrace.pagination import PaginatedList
from dynatrace.utils import timestamp_to_string, iso8601_to_datetime


SCOPE_INSTALLER_DOWNLOAD = "InstallerDownload"  # PaaS integration - Installer download.
SCOPE_DATA_EXPORT = "DataExport"  # Access problem and event feed, metrics, and topology.
SCOPE_PLUGIN_UPLOAD = "PluginUpload"  # Upload Extension.
SCOPE_SUPPORT_ALERT = "SupportAlert"  # PaaS integration - Support alert.
SCOPE_DCRUM_INTEGRATION = "DcrumIntegration"  # Dynatrace module integration - NAM.
SCOPE_ADVANCED_SYNTHETIC_INTEGRATION = "AdvancedSyntheticIntegration"  # Dynatrace module integration - Synthetic Classic.
SCOPE_EXTERNAL_SYNTHETIC_INTEGRATION = "ExternalSyntheticIntegration"  # Create and read synthetic monitors, locations, and nodes.
SCOPE_APP_MON_INTEGRATION = "AppMonIntegration"  # Dynatrace module integration - AppMon.
SCOPE_LOG_EXPORT = "LogExport"  # Read log content.
SCOPE_READ_CONFIG = "ReadConfig"  # Read configuration.
SCOPE_WRITE_CONFIG = "WriteConfig"  # Write configuration.
SCOPE_DTAQLAccess = "DTAQLAccess"  # User sessions.
SCOPE_USER_SESSION_ANONYMIZATION = "UserSessionAnonymization"  # Anonymize user session data for data privacy reasons.
SCOPE_DATA_PRIVACY = "DataPrivacy"  # Change data privacy settings.
SCOPE_CAPTURE_REQUEST_DATA = "CaptureRequestData"  # Capture request data.
SCOPE_DAVIS = "Davis"  # Dynatrace module integration - Davis.
SCOPE_DSS_FILE_MANAGEMENT = "DssFileManagement"  # Mobile symbolication file management.
SCOPE_LOG_IMPORT = "LogImport"  # Log import.
SCOPE_RUM_JAVA_SCRIPT_TAG_MANAGEMENT = "RumJavaScriptTagManagement"  # Real user monitoring JavaScript tag management.
SCOPE_TENANT_TOKEN_MANAGEMENT = "TenantTokenManagement"  # Token management.
SCOPE_ACTIVEGATE_CERT_MANAGEMENT = "ActiveGateCertManagement"  # ActiveGate certificate management.
SCOPE_REST_REQUEST_FORWARDING = "RestRequestForwarding"  # Fetch data from a remote environment.
SCOPE_READ_SYNTHETIC_DATA = "ReadSyntheticData"  # Read synthetic monitors, locations, and nodes.
SCOPE_DATA_IMPORT = "DataImport"  # Data ingest, e.g.  # metrics and events.
SCOPE_AUDIT_LOG_READ = "auditLogs.read"  # Read audit logs.
SCOPE_METRICS_READ = "metrics.read"  # Read metrics.
SCOPE_METRICS_WRITE = "metrics.write"  # Write metrics.
SCOPE_ENTITIES_READ = "entities.read"  # Read entities.
SCOPE_ENTITIES_WRITE = "entities.write"  # Write entities.
SCOPE_PROBLEMS_READ = "problems.read"  # Read problems.
SCOPE_PROBLEMS_WRITE = "problems.write"  # Write problems.
SCOPE_NETWORK_ZONES_READ = "networkZones.read"  # Read network zones.
SCOPE_NETWORK_ZONES_WRITE = "networkZones.write"  # Write network zones.
SCOPE_ACTIVE_GATES_READ = "activeGates.read"  # Read ActiveGates.
SCOPE_ACTIVE_GATES_WRITE = "activeGates.write"  # Write ActiveGates.
SCOPE_CREDENTIAL_VAULT_READ = "credentialVault.read"  # Read credential vault entries.
SCOPE_CREDENTIAL_VAULT_WRITE = "credentialVault.write"  # Write credential vault entries.
SCOPE_EXTENSIONS_READ = "extensions.read"  # Read extensions.
SCOPE_EXTENSIONS_WRITE = "extensions.write"  # Write extensions.
SCOPE_EXTENSION_CONFIGURATIONS_READ = "extensionConfigurations.read"  # Read extension monitoring configurations.
SCOPE_EXTENSION_CONFIGURATIONS_WRITE = "extensionConfigurations.write"  # Write extension monitoring configurations.
SCOPE_EXTENSION_ENVIRONMENT_READ = "extensionEnvironment.read"  # Read extension environment configurations.
SCOPE_EXTENSION_ENVIRONMENT_WRITE = "extensionEnvironment.write"  # Write extension environment configurations.
SCOPE_METRICS_INGEST = "metrics.ingest"  # Ingest metrics.
SCOPE_SECURITY_PROBLEMS_READ = "securityProblems.read"  # Read security problems.
SCOPE_SECURITY_PROBLEMS_WRITE = "securityProblems.write"  # Write security problems.
SCOPE_SYNTHETIC_LOCATIONS_READ = "syntheticLocations.read"  # Read synthetic locations.
SCOPE_SYNTHETIC_LOCATIONS_WRITE = "syntheticLocations.write"  # Write synthetic locations.
SCOPE_TENANT_TOKEN_ROTATION_WRITE = "tenantTokenRotation.write"  # Tenant token rotation.
SCOPE_SLO_READ = "slo.read"  # Read SLO.
SCOPE_SLO_WRITE = "slo.write"  # Write SLO.
SCOPE_RELEASES_READ = "releases.read"  # Read releases.
SCOPE_API_TOKENS_READ = "apiTokens.read"  # Read API tokens.
SCOPE_API_TOKENS_WRITE = "apiTokens.write"  # Write API tokens.


class TokenService:
    def __init__(self, http_client: HttpClient):
        self.__http_client = http_client

    def list(
        self,
        api_token_selector: Optional[str] = None,
        fields: Optional[str] = None,
        time_from: Optional[Union[str, datetime]] = None,
        time_to: Optional[Union[str, datetime]] = None,
        sort: Optional[str] = None,
    ) -> PaginatedList["ApiToken"]:

        params = {
            "apiTokenSelector": api_token_selector,
            "fields": fields,
            "from": timestamp_to_string(time_from),
            "to": timestamp_to_string(time_to),
            "sort": sort,
        }

        return PaginatedList(ApiToken, self.__http_client, "/api/v2/apiTokens", params, list_item="apiTokens")

    def get(self, token_id: str):
        response = self.__http_client.make_request(f"/api/v2/apiTokens/{token_id}", method="GET")
        return ApiToken(raw_element=response.json())

    def delete(self, token_id: str) -> Response:
        return self.__http_client.make_request(f"/api/v2/apiTokens/{token_id}", method="DELETE")

    def create(
        self, name: str, scopes: List[str], personal_access_token: Optional[bool] = None, expiration_date: Optional[Union[datetime, str]] = None
    ) -> "ApiTokenCreated":

        body = {"personalAccessToken": personal_access_token, "expirationDate": expiration_date, "name": name, "scopes": scopes}

        response = self.__http_client.make_request("/api/v2/apiTokens", method="POST", params=body)
        return ApiTokenCreated(raw_element=response.json())

    def lookup(self, token: str):
        response = self.__http_client.make_request("/api/v2/apiTokens/lookup", method="POST", params={"token": token})
        return ApiToken(raw_element=response.json())


class ApiToken(DynatraceObject):
    def _create_from_raw_data(self, raw_element):
        self.last_used_date: Optional[datetime] = iso8601_to_datetime(raw_element.get("lastUsedDate"))
        self.last_used_ip_address: str = raw_element.get("lastUsedIpAddress")
        self.personal_access_token: bool = raw_element.get("personalAccessToken", False)
        self.expiration_date: Optional[datetime] = iso8601_to_datetime(raw_element.get("expirationDate"))
        self.creation_date: Optional[datetime] = iso8601_to_datetime(raw_element.get("creationDate"))
        self.name: str = raw_element.get("name")
        self.id: str = raw_element.get("id")
        self.owner: str = raw_element.get("owner")
        self.scopes: List[str] = raw_element.get("scopes", [])
        self.enabled: bool = raw_element.get("enabled")

    def delete(self):
        return self._http_client.make_request(f"/api/v2/apiTokens/{self.id}", method="DELETE")

    def __repr__(self):
        return f"ApiToken(id={self.id}, owner={self.owner}, name={self.name})"


class ApiTokenCreated(DynatraceObject):
    def _create_from_raw_data(self, raw_element):
        self.token: str = raw_element.get("token")
        self.expiration_date: Optional[datetime] = iso8601_to_datetime(raw_element.get("expirationDate"))
        self.id: str = raw_element.get("id")



from collections import defaultdict
from datetime import datetime
from typing import List, Optional, Dict

from dynatrace.dynatrace_object import DynatraceObject
from dynatrace.http_client import HttpClient

SYNTHETIC_EVENT_TYPE_OUTAGE = "testOutage"
SYNTHETIC_EVENT_TYPE_SLOWDOWN = "testSlowdown"


class ThirdPartySyntheticTestsService:
    def __init__(self, http_client: HttpClient):
        self.__http_client = http_client
        self.__open_third_party_events: Dict[str, int] = defaultdict(int)

    def report_simple_thirdparty_synthetic_test(
        self,
        engine_name: str,
        timestamp: datetime,
        location_id: str,
        location_name: str,
        test_id: str,
        test_title: str,
        schedule_interval: int,
        success: bool,
        response_time: int,
        icon_url: str = None,
        edit_link: str = None,
        step_title: Optional[str] = None,
        detailed_steps: Optional[List["SyntheticTestStep"]] = None,
        detailed_step_results: Optional[List["SyntheticMonitorStepResult"]] = None,
    ):

        location = ThirdPartySyntheticLocation(self.__http_client, location_id, location_name)
        synthetic_location = SyntheticTestLocation(self.__http_client, location_id)
        if detailed_steps is None:
            detailed_steps = [SyntheticTestStep(self.__http_client, 1, step_title)]

        monitor = ThirdPartySyntheticMonitor(
            self.__http_client,
            test_id,
            test_title,
            [synthetic_location],
            schedule_interval,
            steps=detailed_steps,
            edit_link=edit_link,
        )
        if detailed_step_results is None:
            detailed_step_results = [SyntheticMonitorStepResult(self.__http_client, 1, timestamp, response_time_millis=response_time)]
        location_result = ThirdPartySyntheticLocationTestResult(self.__http_client, location_id, timestamp, success, step_results=detailed_step_results)
        test_result = ThirdPartySyntheticResult(self.__http_client, test_id, len(detailed_steps), [location_result])
        tests = ThirdPartySyntheticTests(self.__http_client, engine_name, timestamp, [location], [monitor], [test_result], synthetic_engine_icon_url=icon_url)
        return tests.post()

    def create_synthetic_test_step_result(self, step_id: int, timestamp: datetime, response_time: int) -> "SyntheticMonitorStepResult":
        return SyntheticMonitorStepResult(self.__http_client, step_id, timestamp, response_time_millis=response_time)

    def create_synthetic_test_step(self, step_id: int, step_title: str) -> "SyntheticTestStep":
        return SyntheticTestStep(self.__http_client, step_id, step_title)

    def report_simple_thirdparty_synthetic_test_event(
        self,
        test_id: str,
        name: str,
        location_id: str,
        timestamp: datetime,
        state: str,
        event_type: str,
        reason: str,
        engine_name: str,
    ):
        opened_events: List[ThirdPartyEventOpenNotification] = []
        resolved_events = []
        if state == "open":
            self.__open_third_party_events[test_id] += 1
            event_id = f"{test_id}_{self.__open_third_party_events[test_id]}"

            opened_events.append(ThirdPartyEventOpenNotification(self.__http_client, test_id, event_id, name, event_type, reason, timestamp, [location_id]))
        else:
            if test_id in self.__open_third_party_events:
                event_ids = [f"{test_id}_{i + 1}" for i in range(self.__open_third_party_events[test_id])]
                for event_id in event_ids:
                    resolved_events.append(ThirdPartyEventResolvedNotification(self.__http_client, test_id, event_id, timestamp))
                del self.__open_third_party_events[test_id]

        if opened_events or resolved_events:
            events = ThirdPartySyntheticEvents(self.__http_client, engine_name, opened_events, resolved_events)
            return events.post()


class ThirdPartySyntheticTests(DynatraceObject):
    def __init__(
        self,
        http_client,
        synthetic_engine_name: str,
        message_timestamp: datetime,
        locations: List["ThirdPartySyntheticLocation"],
        tests: List["ThirdPartySyntheticMonitor"],
        test_results: Optional[List["ThirdPartySyntheticResult"]] = None,
        synthetic_engine_icon_url: Optional[str] = None,
    ):

        raw_element = {
            "syntheticEngineName": synthetic_engine_name,
            "syntheticEngineIconUrl": synthetic_engine_icon_url,
            "messageTimestamp": int(message_timestamp.timestamp() * 1000),
            "locations": [location._raw_element for location in locations],
            "tests": [test._raw_element for test in tests],
            "testResults": [test_result._raw_element for test_result in test_results] if test_results else None,
        }
        super().__init__(http_client, None, raw_element)

    def post(self):
        return self._http_client.make_request(f"/api/v1/synthetic/ext/tests", params=self._raw_element, method="POST")


class ThirdPartySyntheticLocation(DynatraceObject):
    def __init__(self, http_client, location_id: str, name: str, ip: Optional[str] = None):

        raw_element = {"id": location_id, "name": name, "ip": ip}
        super().__init__(http_client, None, raw_element)


class ThirdPartySyntheticMonitor(DynatraceObject):
    def __init__(
        self,
        http_client,
        test_id: str,
        title: str,
        locations: List["SyntheticTestLocation"],
        schedule_interval_in_seconds: int,
        description: Optional[str] = None,
        test_setup: Optional[str] = None,
        expiration_timestamp: Optional[int] = None,
        drilldown_link: Optional[str] = None,
        edit_link: Optional[str] = None,
        deleted: Optional[bool] = None,
        steps: Optional[List["SyntheticTestStep"]] = None,
        no_data_timeout: Optional[int] = None,
    ):

        raw_element = {
            "id": test_id,
            "title": title,
            "description": description,
            "testSetup": test_setup,
            "expirationTimestamp": expiration_timestamp,
            "drilldownLink": drilldown_link,
            "editLink": edit_link,
            "deleted": deleted,
            "locations": [location._raw_element for location in locations],
            "steps": [step._raw_element for step in steps] if steps else None,
            "scheduleIntervalInSeconds": schedule_interval_in_seconds,
            "noDataTimeout": no_data_timeout,
        }
        super().__init__(http_client, None, raw_element)


class SyntheticTestLocation(DynatraceObject):
    def __init__(self, http_client, location_id: str, enabled: Optional[bool] = None):

        raw_element = {"id": location_id, "enabled": enabled}
        super().__init__(http_client, None, raw_element)


class SyntheticTestStep(DynatraceObject):
    def __init__(self, http_client, step_id: int, title: str):
        self.step_id = step_id
        self.title = title

        raw_element = {"id": step_id, "title": title}
        super().__init__(http_client, None, raw_element)


class ThirdPartySyntheticResult(DynatraceObject):
    def __init__(
        self,
        http_client,
        test_id: str,
        total_step_count: int,
        location_results: List["ThirdPartySyntheticLocationTestResult"],
        schedule_interval_in_seconds: Optional[int] = None,
    ):

        raw_element = {
            "id": test_id,
            "scheduleIntervalInSeconds": schedule_interval_in_seconds,
            "totalStepCount": total_step_count,
            "locationResults": [location_result._raw_element for location_result in location_results],
        }
        super().__init__(http_client, None, raw_element)


class ThirdPartySyntheticLocationTestResult(DynatraceObject):
    def __init__(
        self,
        http_client,
        location_id: str,
        start_timestamp: datetime,
        success: bool,
        success_rate: Optional[float] = None,
        response_time_millis: Optional[int] = None,
        step_results: Optional[List["SyntheticMonitorStepResult"]] = None,
    ):

        raw_element = {
            "id": location_id,
            "startTimestamp": int(start_timestamp.timestamp() * 1000),
            "successRate": success_rate,
            "success": success,
            "responseTimeMillis": response_time_millis,
            "stepResults": [step_result._raw_element for step_result in step_results] if step_results else None,
        }
        super().__init__(http_client, None, raw_element)


class SyntheticMonitorStepResult(DynatraceObject):
    def __init__(
        self,
        http_client,
        step_id: int,
        start_timestamp: datetime,
        response_time_millis: Optional[int] = None,
        error: Optional["SyntheticMonitorError"] = None,
    ):

        raw_element = {
            "id": step_id,
            "startTimestamp": int(start_timestamp.timestamp() * 1000),
            "responseTimeMillis": response_time_millis,
            "error": error._raw_element if error else None,
        }
        super().__init__(http_client, None, raw_element)


class SyntheticMonitorError(DynatraceObject):
    def __init__(self, http_client, code: int, message: str):

        raw_element = {"code": code, "message": message}
        super().__init__(http_client, None, raw_element)


class ThirdPartySyntheticEvents(DynatraceObject):
    def __init__(
        self,
        http_client,
        synthetic_engine_name: str,
        open_events: Optional[List["ThirdPartyEventOpenNotification"]],
        resolved_events: Optional[List["ThirdPartyEventResolvedNotification"]],
    ):

        raw_element = {
            "syntheticEngineName": synthetic_engine_name,
            "open": [open_event._raw_element for open_event in open_events] if open_events else None,
            "resolved": [resolved_event._raw_element for resolved_event in resolved_events] if resolved_events else None,
        }
        super().__init__(http_client, None, raw_element)

    def post(self):
        return self._http_client.make_request(f"/api/v1/synthetic/ext/events", params=self._raw_element, method="POST")


class ThirdPartyEventOpenNotification(DynatraceObject):
    def __init__(
        self,
        http_client,
        test_id: str,
        event_id: str,
        name: str,
        event_type: str,
        reason: str,
        start_timestamp: datetime,
        location_ids: List[str],
    ):

        raw_element = {
            "testId": test_id,
            "eventId": event_id,
            "name": name,
            "eventType": event_type,
            "reason": reason,
            "startTimestamp": int(start_timestamp.timestamp() * 1000),
            "locationIds": location_ids,
        }
        super().__init__(http_client, None, raw_element)


class ThirdPartyEventResolvedNotification(DynatraceObject):
    def __init__(self, http_client, test_id: str, event_id: str, end_timestamp: datetime):

        raw_element = {"testId": test_id, "eventId": event_id, "endTimestamp": int(end_timestamp.timestamp() * 1000)}
        super().__init__(http_client, None, raw_element)



from dynatrace.dynatrace_object import DynatraceObject


class Tile(DynatraceObject):
    def _create_from_raw_data(self, raw_element):
        self._name = raw_element.get("name")
        self._tile_type = raw_element.get("tileType")  # TODO - This branches, maybe Tile is a subclass
        self._configured = raw_element.get("configured")
        self._bounds = raw_element.get("bounds")  # TODO - This is a TileBounds
        self._tile_filter = raw_element.get("tileFilter")  # TODO - This is a TileFilter



from typing import List, Optional
from requests import Response

from dynatrace.dynatrace_object import DynatraceObject
from dynatrace.environment_v2.entity import EntityShortRepresentation
from dynatrace.http_client import HttpClient
from dynatrace.pagination import PaginatedList
from dynatrace.configuration_v1.tile import Tile


class DashboardService:
    def __init__(self, http_client: HttpClient):
        self.__http_client = http_client

    def list(self, owner: str = None, tags: List[str] = None) -> PaginatedList["DashboardStub"]:
        """
        Lists all dashboards of the environment
        :param owner: The owner of the dashboard.
        :param tags: A list of tags applied to the dashboard.
            The dashboard must match all the specified tags.
        """
        params = {"owner": owner, "tags": tags}
        return PaginatedList(DashboardStub, self.__http_client, f"/api/config/v1/dashboards", params, list_item="dashboards")

    def get(self, dashboard_id: str) -> "Dashboard":
        """
        Gets the properties of the specified dashboard
        """
        response = self.__http_client.make_request(f"/api/config/v1/dashboards/{dashboard_id}").json()
        return Dashboard(self.__http_client, None, response)

    def delete(self, dashboard_id: str) -> Response:
        """
        Deletes the specified dashboard
        """
        return self.__http_client.make_request(f"/api/config/v1/dashboards/{dashboard_id}", method="DELETE")


class DashboardFilter(DynatraceObject):
    @property
    def timeframe(self) -> str:
        return self._timeframe

    @property
    def management_zone(self) -> Optional[EntityShortRepresentation]:
        return self._management_zone

    def _create_from_raw_data(self, raw_element):
        if raw_element is None:
            raw_element = {}
        self._timeframe = raw_element.get("timeframe")
        self._management_zone = (
            EntityShortRepresentation(self._http_client, None, raw_element.get("managementZone")) if raw_element.get("managementZone") else None
        )


class SharingInfo(DynatraceObject):
    @property
    def link_shared(self):
        return self._link_shared

    @property
    def published(self):
        return self.published

    def _create_from_raw_data(self, raw_element):
        self._link_shared = raw_element.get("linkShared")
        self._published = raw_element.get("published")


class DashboardMetadata(DynatraceObject):
    @property
    def name(self) -> str:
        return self._name

    @property
    def shared(self) -> bool:
        return self._shared

    @property
    def owner(self) -> str:
        return self._owner

    @property
    def dashboard_filter(self):
        return self._dashboard_filter

    @property
    def tags(self) -> List[str]:
        return self._tags

    @property
    def preset(self) -> bool:
        return self._preset

    def _create_from_raw_data(self, raw_element):
        # TODO - Schema changed, add extra fields dynamicFilters
        self._name = raw_element.get("name")
        self._shared = raw_element.get("shared")
        self._owner = raw_element.get("owner")
        self._dashboard_filter = DashboardFilter(self._http_client, None, raw_element.get("dashboardFilter"))
        self._tags = raw_element.get("tags")
        self._preset = raw_element.get("preset")


class Dashboard(DynatraceObject):
    @property
    def id(self) -> str:
        return self._id

    @property
    def dashboard_metadata(self) -> DashboardMetadata:
        return self._dashboard_metadata

    @property
    def tiles(self) -> List[Tile]:
        return self._tiles

    def _create_from_raw_data(self, raw_element):
        if raw_element is None:
            raw_element = {}
        self._id = raw_element.get("id")
        self._dashboard_metadata = DashboardMetadata(self._http_client, None, raw_element.get("dashboardMetadata"))
        self._tiles = [Tile(self._http_client, None, raw_tile) for raw_tile in raw_element.get("tiles", [])]


class DashboardStub(DynatraceObject):
    @property
    def id(self) -> str:
        return self._id

    @property
    def name(self) -> str:
        return self._name

    @property
    def owner(self) -> str:
        return self._owner

    def delete(self) -> Response:
        """
        Deletes this dashboard
        """
        return self._http_client.make_request(f"/api/config/v1/dashboards/{self.id}", method="DELETE")

    def _create_from_raw_data(self, raw_element):
        self._id = raw_element.get("id")
        self._name = raw_element.get("name")
        self._owner = raw_element.get("owner")

    def get_full_dashboard(self) -> Dashboard:
        """
        Gets the full dashboard for this stub
        """
        response = self._http_client.make_request(f"/api/config/v1/dashboards/{self.id}").json()
        return Dashboard(self._http_client, None, response)



from typing import List, Optional

from dynatrace.dynatrace_object import DynatraceObject
from dynatrace.http_client import HttpClient
from dynatrace.pagination import PaginatedList

OS_TYPE_LINUX = "LINUX"
OS_TYPE_WINDOWS = "WINDOWS"

TYPE_ENVIRONMENT = "ENVIRONMENT"
TYPE_ENVIRONMENT_MULTI = "ENVIRONMENT_MULTI"

UPDATE_STATUS_INCOMPATIBLE = "INCOMPATIBLE"
UPDATE_STATUS_OUTDATED = "OUTDATED"
UPDATE_STATUS_SUPPRESSED = "SUPPRESSED"
UPDATE_STATUS_UNKNOWN = "UNKNOWN"
UPDATE_STATUS_UP2DATE = "UP2DATE"
UPDATE_STATUS_UPDATE_IN_PROGRESS = "UPDATE_IN_PROGRESS"
UPDATE_STATUS_UPDATE_PENDING = "UPDATE_PENDING"
UPDATE_STATUS_UPDATE_PROBLEM = "UPDATE_PROBLEM"

VERSION_COMPARE_TYPE_EQUAL = "EQUAL"
VERSION_COMPARE_TYPE_GREATER = "GREATER"
VERSION_COMPARE_TYPE_GREATER_EQUAL = "GREATER_EQUAL"
VERSION_COMPARE_TYPE_LOWER = "LOWER"
VERSION_COMPARE_TYPE_LOWER_EQUAL = "LOWER_EQUAL"

AUTO_UPDATE_ENABLED = "ENABLED"
AUTO_UPDATE_DISABLED = "DISABLED"

MODULE_AWS = "AWS"
MODULE_AZURE = "AZURE"
MODULE_BEACON_FORWARDER = "BEACON_FORWARDER"
MODULE_CLOUD_FOUNDRY = "CLOUD_FOUNDRY"
MODULE_DB_INSIGHT = "DB_INSIGHT"
MODULE_EXTENSIONS_V1 = "EXTENSIONS_V1"
MODULE_EXTENSIONS_V2 = "EXTENSIONS_V2"
MODULE_KUBERNETES = "KUBERNETES"
MODULE_LOGS = "LOGS"
MODULE_MEMORY_DUMPS = "MEMORY_DUMPS"
MODULE_METRIC_API = "METRIC_API"
MODULE_ONE_AGENT_ROUTING = "ONE_AGENT_ROUTING"
MODULE_OTLP_INGEST = "OTLP_INGEST"
MODULE_REST_API = "REST_API"
MODULE_SYNTHETIC = "SYNTHETIC"


class ActiveGateService:
    def __init__(self, http_client: HttpClient):
        self.__http_client = http_client

    def list(
        self,
        hostname: Optional[str] = None,
        os_type: Optional[str] = None,
        network_address: Optional[str] = None,
        activegate_type: Optional[str] = None,
        network_zone: Optional[str] = None,
        update_status: Optional[str] = None,
        version_compare_type: Optional[str] = None,
        version: Optional[str] = None,
        auto_update: Optional[str] = None,
        group: Optional[str] = None,
        online: Optional[bool] = None,
        enabled_modules: Optional[List[str]] = None,
        disabled_modules: Optional[List[str]] = None,
    ) -> PaginatedList["ActiveGate"]:
        """
        Lists all available ActiveGates
        :return: A list of ActiveGates.
        """
        params = {
            "hostname": hostname,
            "osType": os_type,
            "networkAddress": network_address,
            "ActivegateType": activegate_type,
            "networkZone": network_zone,
            "updateStatus": update_status,
            "versionCompareType": version_compare_type,
            "version": version,
            "autoUpdate": auto_update,
            "group": group,
            "online": online,
            "enabledModule": enabled_modules,
            "disabledModule": disabled_modules,
        }
        return PaginatedList(ActiveGate, self.__http_client, "/api/v2/activeGates", params, list_item="activeGates")

    def get(self, activegate_id: str) -> "ActiveGate":
        return ActiveGate(raw_element=self.__http_client.make_request(f"/api/v2/activeGates/{activegate_id}").json())


class ActiveGate(DynatraceObject):
    def _create_from_raw_data(self, raw_element: dict):
        self.id: str = raw_element.get("id")
        self.network_addresses = raw_element.get("networkAddresses", [])
        self.os_type: str = raw_element.get("osType")
        self.auto_update_status: str = raw_element.get("autoUpdateStatus")
        self.offline_since: int = raw_element.get("offlineSince")
        self.version: str = raw_element.get("version")
        self.type: str = raw_element.get("type")
        self.hostname: str = raw_element.get("hostname")
        self.main_environment: str = raw_element.get("mainEnvironment")
        self.environments: str = raw_element.get("environments", [])
        self.network_zone: str = raw_element.get("networkZone")
        self.group: str = raw_element.get("group")
        self.modules: List["ActiveGateModule"] = [ActiveGateModule(raw_element=module) for module in raw_element.get("modules")]


class ActiveGateModule(DynatraceObject):
    def _create_from_raw_data(self, raw_element: dict):
        self.misconfigured: bool = raw_element.get("misconfigured")
        self.type: str = raw_element.get("type")
        self.attributes: bool = raw_element.get("attributes")
        self.enabled: bool = raw_element.get("enabled")
        self.version: str = raw_element.get("version")



from datetime import datetime

from requests import Response
from dynatrace.environment_v2.entity import EntityShortRepresentation
from dynatrace.configuration_v1.endpoint import EndpointShortRepresentation
from dynatrace.http_client import HttpClient
from dynatrace.pagination import PaginatedList
from dynatrace.dynatrace_object import DynatraceObject


class PluginService:
    def __init__(self, http_client: HttpClient):
        self.__http_client = http_client

    def list(self) -> PaginatedList["PluginShortRepresentation"]:
        """
        List all uploaded plugins
        """
        return PaginatedList(PluginShortRepresentation, self.__http_client, "/api/config/v1/plugins", list_item="values")

    def list_states(self, plugin_id) -> PaginatedList["PluginState"]:
        """
        List the states of the specified plugin
        """
        return PaginatedList(PluginState, self.__http_client, f"/api/config/v1/plugins/{plugin_id}/states", list_item="states")

    def delete(self, plugin_id) -> Response:
        """
        Deletes the ZIP file of the specified plugin
        :param plugin_id: The ID of the plugin to be deleted
        """
        return self.__http_client.make_request(f"/api/config/v1/plugins/{plugin_id}/binary", method="DELETE")


class PluginState(DynatraceObject):
    @property
    def plugin_id(self) -> str:
        """
        The ID of the plugin.
        """
        return self._plugin_id

    @property
    def version(self) -> str:
        """
        The version of the plugin (for example 1.0.0).
        """
        return self._version

    @property
    def endpoint_id(self) -> str:
        """
        The ID of the endpoint where the state is detected - Active Gate only.
        """
        return self._endpoint_id

    @property
    def state(self) -> str:
        """
        The state of the plugin.
        [ DISABLED, ERROR_AUTH, ERROR_COMMUNICATION_FAILURE, ERROR_CONFIG, ERROR_TIMEOUT, ERROR_UNKNOWN,
         INCOMPATIBLE, LIMIT_REACHED, NOTHING_TO_REPORT, OK, STATE_TYPE_UNKNOWN, UNINITIALIZED, UNSUPPORTED, WAITING_FOR_STATE ]
        """
        return self._state

    @property
    def state_description(self) -> str:
        """
        A short description of the state.
        """
        return self._state_description

    @property
    def timestamp(self) -> datetime:
        """
        The timestamp when the state was detected
        """
        return datetime.utcfromtimestamp(self._timestamp / 1000)

    @property
    def host_id(self) -> str:
        """
        The ID of the host on which the plugin runs.
        """
        return self._host_id

    @property
    def process_id(self) -> str:
        """
        The ID of the entity on which the plugin is active.
        """
        return self._process_id

    def _create_from_raw_data(self, raw_element):
        self._plugin_id = raw_element.get("pluginId")
        self._version = raw_element.get("version")
        self._endpoint_id = raw_element.get("endpointId")
        self._state = raw_element.get("state")
        self._state_description = raw_element.get("stateDescription")
        self._timestamp = raw_element.get("timestamp")
        self._host_id = raw_element.get("hostId")
        self._process_id = raw_element.get("processId")


class PluginShortRepresentation(EntityShortRepresentation):
    def delete(self) -> Response:
        """
        Deletes the ZIP file of this plugin
        """
        return self._http_client.make_request(f"/api/config/v1/plugins/{self.id}/binary", method="DELETE")

    @property
    def endpoints(self) -> PaginatedList[EndpointShortRepresentation]:
        return PaginatedList(
            EndpointShortRepresentation,
            self._http_client,
            f"/api/config/v1/plugins/{self.id}/endpoints",
            None,
            list_item="values",
        )

    @property
    def states(self) -> PaginatedList[PluginState]:
        return PaginatedList(PluginState, self._http_client, f"/api/config/v1/plugins/{self.id}/states", None, list_item="states")



from dynatrace.dynatrace_object import DynatraceObject
from dynatrace.environment_v2.configuration import ConfigurationMetadata
from dynatrace.http_client import HttpClient


class ActiveGateAutoUpdateService:
    def __init__(self, http_client: HttpClient):
        self.__http_client = http_client

    def get_global(self) -> "ActiveGateGlobalAutoUpdateConfig":
        return ActiveGateGlobalAutoUpdateConfig(raw_element=self.__http_client.make_request("/api/v2/activeGates/autoUpdate").json())


class ActiveGateGlobalAutoUpdateConfig(DynatraceObject):
    def _create_from_raw_data(self, raw_element):
        self.global_setting: str = raw_element.get("globalSetting")
        self.metadata: ConfigurationMetadata = ConfigurationMetadata(raw_element=raw_element.get("metadata"))



import logging
from typing import Dict, Optional

from dynatrace.configuration_v1.dashboard import DashboardService
from dynatrace.configuration_v1.plugins import PluginService
from dynatrace.environment_v1.cluster_time import ClusterTimeService
from dynatrace.environment_v1.custom_device import CustomDeviceService
from dynatrace.environment_v1.event import EventService
from dynatrace.environment_v1.synthetic_third_party import ThirdPartySyntheticTestsService
from dynatrace.environment_v2.activegate import ActiveGateService
from dynatrace.environment_v2.activegate_autoupdate import ActiveGateAutoUpdateService
from dynatrace.environment_v2.entity import EntityService
from dynatrace.environment_v2.extension import ExtensionService
from dynatrace.environment_v2.metric import MetricService
from dynatrace.environment_v2.token import TokenService
from dynatrace.http_client import HttpClient


class Dynatrace:
    def __init__(
        self,
        base_url: str,
        token: str,
        log: logging.Logger = None,
        proxies: Dict = None,
        too_many_requests_strategy=None,
        retries: int = 0,
        retry_delay_ms: int = 0,
        mc_jsession_id: Optional[str] = None,
        mc_b925d32c: Optional[str] = None,
        mc_sso_csrf_cookie: Optional[str] = None,
    ):
        self.__http_client = HttpClient(
            base_url,
            token,
            log,
            proxies,
            too_many_requests_strategy,
            retries,
            retry_delay_ms,
            mc_jsession_id,
            mc_b925d32c,
            mc_sso_csrf_cookie,
        )

        self.activegates: ActiveGateService = ActiveGateService(self.__http_client)
        self.activegates_autoupdate: ActiveGateAutoUpdateService = ActiveGateAutoUpdateService(self.__http_client)
        self.cluster_time: ClusterTimeService = ClusterTimeService(self.__http_client)
        self.custom_devices: CustomDeviceService = CustomDeviceService(self.__http_client)
        self.dashboards: DashboardService = DashboardService(self.__http_client)
        self.entities: EntityService = EntityService(self.__http_client)
        self.events: EventService = EventService(self.__http_client)
        self.extensions: ExtensionService = ExtensionService(self.__http_client)
        self.metrics: MetricService = MetricService(self.__http_client)
        self.plugins: PluginService = PluginService(self.__http_client)
        self.third_part_synthetic_tests: ThirdPartySyntheticTestsService = ThirdPartySyntheticTestsService(self.__http_client)
        self.tokens: TokenService = TokenService(self.__http_client)



import warnings
import functools
from datetime import datetime
from typing import Union, Optional

ISO_8601 = "%Y-%m-%dT%H:%M:%S.%fZ"


def deprecated(reason=""):
    def decorator(func):
        @functools.wraps(func)
        def new_func(*args, **kwargs):
            warnings.warn(f"'{func.__name__}' is deprecated. {reason}", category=DeprecationWarning, stacklevel=2)
            return func(*args, **kwargs)

        return new_func

    return decorator


def timestamp_to_string(timestamp: Optional[Union[datetime, str]]) -> Optional[str]:
    if not isinstance(timestamp, datetime):
        return timestamp
    return timestamp.isoformat(timespec="milliseconds")


def iso8601_to_datetime(timestamp: Optional[str]) -> Optional[datetime]:
    if isinstance(timestamp, str):
        return datetime.strptime(timestamp, ISO_8601)
    return timestamp



import logging
from typing import Dict, Optional
import time

import requests
import urllib3
from urllib3.util.retry import Retry
from requests.adapters import HTTPAdapter

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


TOO_MANY_REQUESTS_WAIT = "wait"


class DynatraceRetry(Retry):
    def get_backoff_time(self):
        return self.backoff_factor


class HttpClient:
    def __init__(
        self,
        base_url: str,
        token: str,
        log: logging.Logger = None,
        proxies: Dict = None,
        too_many_requests_strategy=None,
        retries: int = 0,
        retry_delay_ms: int = 0,
        mc_jsession_id: Optional[str] = None,
        mc_b925d32c: Optional[str] = None,
        mc_sso_csrf_cookie: Optional[str] = None,
    ):
        while base_url.endswith("/"):
            base_url = base_url[:-1]
        self.base_url = base_url

        if proxies is None:
            proxies = {}
        self.proxies = proxies

        self.auth_header = {"Authorization": f"Api-Token {token}"}
        self.log = log
        if self.log is None:
            self.log = logging.getLogger(__name__)
            self.log.setLevel(logging.WARNING)
            st = logging.StreamHandler()
            fmt = logging.Formatter("%(asctime)s - %(levelname)s - %(name)s - %(thread)d - %(filename)s:%(lineno)d - %(message)s")
            st.setFormatter(fmt)
            self.log.addHandler(st)

        self.too_many_requests_strategy = too_many_requests_strategy
        retry_delay_s = retry_delay_ms / 1000
        self.retries = DynatraceRetry(
            total=retries,
            backoff_factor=retry_delay_s,
            status_forcelist=[400, 401, 403, 413, 429, 500, 502, 503, 504],
            method_whitelist=["TRACE", "PUT", "DELETE", "OPTIONS", "HEAD", "GET", "POST"],
            raise_on_status=False,
        )

        # This is for internal dynatrace usage
        self.mc_jsession_id = mc_jsession_id
        self.mc_b925d32c = mc_b925d32c
        self.mc_sso_csrf_cookie = mc_sso_csrf_cookie

    def make_request(self, path: str, params: Optional[Dict] = None, headers: Optional[Dict] = None, method="GET", data=None) -> requests.Response:
        url = f"{self.base_url}{path}"

        body = None
        if method in ["POST", "PUT"]:
            body = params
            params = None

        if headers is None:
            headers = {"content-type": "application/json"}
        headers.update(self.auth_header)

        cookies = None
        if self.mc_b925d32c and self.mc_sso_csrf_cookie and self.mc_jsession_id:
            headers.update({"Cookie": f"JSESSIONID={self.mc_jsession_id}; ssoCSRFCookie={self.mc_sso_csrf_cookie}; b925d32c={self.mc_b925d32c}"})
            cookies = {"JSESSIONID": self.mc_jsession_id, "ssoCSRFCookie": self.mc_sso_csrf_cookie, "b925d32c": self.mc_b925d32c}

        s = requests.Session()
        s.mount("https://", HTTPAdapter(max_retries=self.retries))

        self.log.debug(f"Making {method} request to '{url}' with params {params} and body: {body}")
        r = s.request(method, url, headers=headers, params=params, json=body, verify=False, proxies=self.proxies, data=data, cookies=cookies)
        self.log.debug(f"Received response '{r}'")

        while r.status_code == 429 and self.too_many_requests_strategy == TOO_MANY_REQUESTS_WAIT:
            sleep_amount = int(r.headers.get("retry-after", 5))
            self.log.warning(f"Sleeping for {sleep_amount}s because we have received an HTTP 429")
            time.sleep(sleep_amount)
            r = requests.request(method, url, headers=headers, params=params, json=body, verify=False, proxies=self.proxies)

        if r.status_code >= 400:
            raise Exception(f"Error making request to {url}: {r}. Response: {r.text}, Parameters: {params}, Body: {body}, Headers: {r.headers}")

        return r



from datetime import datetime
from typing import List

from dynatrace.dynatrace_object import DynatraceObject
from dynatrace.http_client import HttpClient
from dynatrace.pagination import PaginatedList


class MetricService:
    def __init__(self, http_client: HttpClient):
        self.__http_client = http_client

    def query(
        self,
        metric_selector: str,
        resolution: str = None,
        time_from=None,
        time_to=None,
        entity_selector=None,
    ) -> PaginatedList["MetricSeriesCollection"]:
        """

        :param metric_selector: The metric selector: https://www.dynatrace.com/support/help/shortlink/api-metrics-v2-selector
        :param resolution:
        :param time_from:
        :param time_to:
        :param entity_selector:
        :return:
        """

        params = {
            "metricSelector": metric_selector,
            "resolution": resolution,
            "from": time_from,
            "to": time_to,
            "entitySelector": entity_selector,
        }
        return PaginatedList(MetricSeriesCollection, self.__http_client, "/api/v2/metrics/query", params, list_item="result")

    def list(self, metric_selector: str = None, text: str = None, fields: str = None, page_size=100) -> PaginatedList["MetricDescriptor"]:
        params = {
            "pageSize": page_size,
            "metricSelector": metric_selector,
            "text": text,
            "fields": fields,
        }
        return PaginatedList(MetricDescriptor, self.__http_client, "/api/v2/metrics", params, list_item="metrics")

    def get(self, metric_id: str) -> "MetricDescriptor":
        response = self.__http_client.make_request(f"/api/v2/metrics/{metric_id}").json()
        return MetricDescriptor(self.__http_client, None, response)

    def ingest(self, lines: List[str]):
        lines = "\n".join(lines)
        return self.__http_client.make_request(
            f"/api/v2/metrics/ingest", method="POST", data=lines, headers={"Content-Type": "text/plain; charset=utf-8"}
        ).json()


class MetricSeries(DynatraceObject):
    @property
    def timestamps(self) -> List[datetime]:
        return self._timestamps

    @property
    def dimensions(self) -> List[str]:
        return self._dimensions

    @property
    def values(self) -> List[float]:
        return self._values

    def _create_from_raw_data(self, raw_element):
        self._timestamps = [datetime.utcfromtimestamp(timestamp / 1000) for timestamp in raw_element.get("timestamps", [])]
        self._dimensions = raw_element.get("dimensions", [])
        self._values = raw_element.get("values", [])


class MetricSeriesCollection(DynatraceObject):
    @property
    def metric_id(self) -> str:
        return self._metric_id

    @property
    def data(self) -> List[MetricSeries]:
        return self._data

    def _create_from_raw_data(self, raw_element: dict):
        self._metric_id = raw_element.get("metricId")
        self._data = [MetricSeries(self._http_client, self._headers, metric_serie) for metric_serie in raw_element.get("data", [])]


class MetricDefaultAggregation(DynatraceObject):
    @property
    def parameter(self) -> float:
        """
        The percentile to be delivered. Valid values are between 0 and 100.
        Applicable only to the percentile aggregation type.
        """
        return self._parameter

    @property
    def type(self) -> str:
        """
        The type of default aggregation.
        [ auto, avg, count, max, median, min, percentile, sum, value ]
        """
        return self._type

    def _create_from_raw_data(self, raw_element):
        self._parameter = raw_element.get("parameter")
        self._type = raw_element.get("type")


class MetricDimensionDefinition(DynatraceObject):
    @property
    def index(self) -> int:
        """
        The unique 0-based index of the dimension.
        :return:
        """
        return self._index

    @property
    def name(self) -> int:
        """
        The name of the dimension.
        :return:
        """
        return self._name

    @property
    def key(self) -> int:
        """
        The key of the dimension.
        :return:
        """
        return self._key

    @property
    def type(self) -> int:
        """
        The type of the dimension.
        :return:[ ENTITY, NUMBER, OTHER, STRING, VOID ]
        """
        return self._type

    def _create_from_raw_data(self, raw_element):
        self._index = raw_element.get("index")
        self._name = raw_element.get("name")
        self._key = raw_element.get("key")
        self._type = raw_element.get("type")


class MetricDescriptor(DynatraceObject):
    @property
    def default_aggregation(self) -> MetricDefaultAggregation:
        """
        The default aggregation of a metric.
        """
        return self._default_aggregation

    @property
    def dimension_definitions(self) -> List[MetricDimensionDefinition]:
        """
        The fine metric division (for example, process group and process ID for some process-related metric).
        """
        return self._dimension_definitions

    @property
    def metric_id(self) -> str:
        """
        The fully qualified key of the metric.
        If a transformation has been used it is reflected in the metric key.
        """
        return self._metric_id

    @property
    def entity_type(self):
        return self._entity_type

    @property
    def aggregation_types(self):
        return self._aggregation_types

    @property
    def description(self):
        return self._description

    @property
    def transformations(self):
        return self._transformations

    @property
    def unit(self):
        return self._unit

    @property
    def display_name(self):
        return self._display_name

    def _create_from_raw_data(self, raw_element):
        self._default_aggregation = (
            MetricDefaultAggregation(self._http_client, self._headers, raw_element.get("defaultAggregation")) if raw_element.get("defaultAggregation") else None
        )
        self._dimension_definitions = [
            MetricDimensionDefinition(self._http_client, self._headers, element) for element in raw_element.get("dimensionDefinitions", [])
        ]

        self._metric_id = raw_element.get("metricId")
        self._description = raw_element.get("description")
        self._entity_type = raw_element.get("entityType")
        self._aggregation_types = raw_element.get("aggregationTypes")
        self._transformations = raw_element.get("transformations")
        self._unit = raw_element.get("unit")
        self._display_name = raw_element.get("displayName")
